import math
from typing import Sequence

import torch

# Physical constants
EP0 = 8.8541878128e-12  # vacuum permittivity
MU0 = 1.2566370614359173e-06  # vacuum permeability
C0 = 1.0 / math.sqrt(EP0 * MU0)  # speed of light in vacuum


def prepare_parameters(
    epsilon_r: torch.Tensor, sigma: torch.Tensor, mu_r: torch.Tensor, dt: float
) -> tuple:
    """Prepare update coefficients for Maxwell equations.

    Args:
        epsilon_r: Relative permittivity.
        sigma: Conductivity (S/m).
        mu_r: Relative permeability.
        dt: Time step (s).

    Returns:
        Tuple of (ca, cb, cq) coefficients.
    """
    # Convert to absolute values
    epsilon = epsilon_r * EP0
    mu = mu_r * MU0

    # Ca and Cb for E-field update: E^{n+1} = Ca*E^n + Cb*(curl H)
    # Ca = (1 - sigma*dt/(2*epsilon)) / (1 + sigma*dt/(2*epsilon))
    # Cb = dt/epsilon / (1 + sigma*dt/(2*epsilon))
    denom = 1.0 + sigma * dt / (2.0 * epsilon)
    ca = (1.0 - sigma * dt / (2.0 * epsilon)) / denom
    cb = (dt / epsilon) / denom

    # Cq for H-field update: H^{n+1/2} = H^{n-1/2} - Cq*(curl E)
    cq = dt / mu

    return ca, cb, cq


def setup_pml(
    pml_width: Sequence[int],
    pml_start: Sequence[float],
    max_pml: float,
    dt: float,
    n: int,
    max_vel: float,
    dtype: torch.dtype,
    device: torch.device,
    pml_freq: float,
    start: float = 0.0,
    r_val: float = 1e-8,
    n_power: int = 4,
    eps: float = 1e-9,
    grid_spacing: float = 1.0,
) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
    """Creates a, b, and k profiles for electromagnetic C-PML.

    This implementation follows the standard CPML formulation for electromagnetic
    wave simulation with proper physical parameters.

    Only the first fd_pad[0]+pml_width[0] and last fd_pad[1]+pml_width[1]
    elements of the profiles will be non-zero.

    Args:
        pml_width: list of two integers specifying the width of the PML
            region [left/bottom, right/top].
        pml_start: list of two floats specifying the coordinates (in grid
            cells) of the start of the PML regions.
        max_pml: Float specifying the length (in distance units) of the
            longest PML over all sides and dimensions.
        dt: Time step interval.
        n: Integer specifying desired profile length, including fd_pad and
           pml_width.
        max_vel: Maximum wave speed (not used in EM formulation, kept for API).
        dtype: PyTorch datatype to use.
        device: PyTorch device to use.
        pml_freq: The frequency value to use for the profile (not used in EM,
            kept for API compatibility).
        start: Float specifying the coordinate (in grid cells) of the first
            element. Optional, default 0.
        r_val: The reflection coefficient. Optional, default 1e-8.
        n_power: The power for the profile. Optional, default 4.
        eps: A small number to prevent division by zero. Optional,
            default 1e-9.
        grid_spacing: The grid spacing (dx or dy). Optional, default 1.0.

    Returns:
        A tuple containing the (a, b, k) profiles as Tensors.
        - a: CPML 'a' coefficient for recursive convolution
        - b: CPML 'b' coefficient for recursive convolution
        - k: CPML stretching factor

    """
    # CPML parameters for electromagnetic waves
    k_max_cpml = 5.0  # maximum stretching factor
    alpha_max_cpml = 0.008  # maximum frequency shift

    # Create output tensors
    a = torch.zeros(n, device=device, dtype=dtype)
    b = torch.zeros(n, device=device, dtype=dtype)
    k = torch.ones(n, device=device, dtype=dtype)

    if max_pml == 0 or (pml_width[0] == 0 and pml_width[1] == 0):
        return a, b, k

    # Standard CPML sigma_max: sig0 = (Npower+1) / (150 * pi * dx)
    sigma0 = (n_power + 1) / (150.0 * math.pi * grid_spacing)

    # Calculate profiles for each grid point
    x = torch.arange(start, start + n, device=device, dtype=dtype)

    # Left/bottom PML region
    if pml_width[0] > 0:
        origin_left = pml_start[0]  # This is the inner edge of PML (in grid cells)
        abscissa_left = origin_left - x  # Distance from inner edge (in grid cells)
        mask_left = abscissa_left >= 0

        # Normalized distance into PML (0 at inner edge, 1 at outer edge)
        # abscissa_left is in grid cells, pml_width[0] is also in grid cells
        abscissa_norm_left = torch.clamp(abscissa_left / pml_width[0], 0, 1)

        # Sigma, k, alpha profiles with polynomial grading
        sigma_left = sigma0 * (abscissa_norm_left**n_power)
        k_left = 1.0 + (k_max_cpml - 1.0) * (abscissa_norm_left**n_power)
        alpha_left = alpha_max_cpml * (1.0 - abscissa_norm_left) + 0.1 * alpha_max_cpml

        # Apply to left region
        k = torch.where(mask_left, k_left, k)

        # Calculate b = exp(-(sigma/k + alpha) * dt / epsilon0)
        b_left = torch.exp(-(sigma_left / k_left + alpha_left) * dt / EP0)
        b = torch.where(mask_left, b_left, b)

        # Calculate a = sigma * (b - 1) / (k * (sigma + k * alpha))
        denom_left = k_left * (sigma_left + k_left * alpha_left) + eps
        a_left = sigma_left * (b_left - 1.0) / denom_left
        # Only apply where sigma is significant
        a_left = torch.where(sigma_left > 1e-6, a_left, torch.zeros_like(a_left))
        a = torch.where(mask_left, a_left, a)

    # Right/top PML region
    if pml_width[1] > 0:
        origin_right = pml_start[1]  # This is the inner edge of PML (in grid cells)
        abscissa_right = x - origin_right  # Distance from inner edge (in grid cells)
        mask_right = abscissa_right >= 0

        # Normalized distance into PML (0 at inner edge, 1 at outer edge)
        # abscissa_right is in grid cells, pml_width[1] is also in grid cells
        abscissa_norm_right = torch.clamp(abscissa_right / pml_width[1], 0, 1)

        # Sigma, k, alpha profiles
        sigma_right = sigma0 * (abscissa_norm_right**n_power)
        k_right = 1.0 + (k_max_cpml - 1.0) * (abscissa_norm_right**n_power)
        alpha_right = (
            alpha_max_cpml * (1.0 - abscissa_norm_right) + 0.1 * alpha_max_cpml
        )

        # Apply to right region
        k = torch.where(mask_right, k_right, k)

        # Calculate b = exp(-(sigma/k + alpha) * dt / epsilon0)
        b_right = torch.exp(-(sigma_right / k_right + alpha_right) * dt / EP0)
        b = torch.where(mask_right, b_right, b)

        # Calculate a = sigma * (b - 1) / (k * (sigma + k * alpha))
        denom_right = k_right * (sigma_right + k_right * alpha_right) + eps
        a_right = sigma_right * (b_right - 1.0) / denom_right
        a_right = torch.where(sigma_right > 1e-6, a_right, torch.zeros_like(a_right))
        a = torch.where(mask_right, a_right, a)

    return a, b, k


def setup_pml_half(
    pml_width: Sequence[int],
    pml_start: Sequence[float],
    max_pml: float,
    dt: float,
    n: int,
    max_vel: float,
    dtype: torch.dtype,
    device: torch.device,
    pml_freq: float,
    start: float = 0.0,
    r_val: float = 1e-8,
    n_power: int = 4,
    eps: float = 1e-9,
    grid_spacing: float = 1.0,
) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
    """Creates a, b, and k profiles for C-PML at half grid points.

    This is used for staggered grid implementations where some field components
    are located at half grid points (e.g., H fields in Yee grid).

    Args:
        Same as setup_pml.

    Returns:
        A tuple containing the (a_half, b_half, k_half) profiles as Tensors.

    """
    # CPML parameters for electromagnetic waves
    k_max_cpml = 5.0
    alpha_max_cpml = 0.008

    a = torch.zeros(n, device=device, dtype=dtype)
    b = torch.zeros(n, device=device, dtype=dtype)
    k = torch.ones(n, device=device, dtype=dtype)

    if max_pml == 0 or (pml_width[0] == 0 and pml_width[1] == 0):
        return a, b, k

    # Standard CPML sigma_max: sig0 = (Npower+1) / (150 * pi * dx)
    sigma0 = (n_power + 1) / (150.0 * math.pi * grid_spacing)

    # Half grid positions (shifted by dx/2 or dy/2)
    x = torch.arange(start, start + n, device=device, dtype=dtype) + 0.5

    # Left/bottom PML region (half grid)
    if pml_width[0] > 0:
        origin_left = pml_start[0]  # Inner edge of PML (in grid cells)
        abscissa_left = origin_left - x  # Distance in grid cells
        mask_left = abscissa_left >= 0

        # Normalized distance (both in grid cells)
        abscissa_norm_left = torch.clamp(abscissa_left / pml_width[0], 0, 1)

        sigma_left = sigma0 * (abscissa_norm_left**n_power)
        k_left = 1.0 + (k_max_cpml - 1.0) * (abscissa_norm_left**n_power)
        alpha_left = alpha_max_cpml * (1.0 - abscissa_norm_left) + 0.1 * alpha_max_cpml

        k = torch.where(mask_left, k_left, k)

        # b = exp(-(sigma/k + alpha) * dt / epsilon0)
        b_left = torch.exp(-(sigma_left / k_left + alpha_left) * dt / EP0)
        b = torch.where(mask_left, b_left, b)

        # a = sigma * (b - 1) / (k * (sigma + k * alpha))
        denom_left = k_left * (sigma_left + k_left * alpha_left) + eps
        a_left = sigma_left * (b_left - 1.0) / denom_left
        a_left = torch.where(sigma_left > 1e-6, a_left, torch.zeros_like(a_left))
        a = torch.where(mask_left, a_left, a)

    # Right/top PML region (half grid)
    if pml_width[1] > 0:
        origin_right = pml_start[1]  # Inner edge of PML (in grid cells)
        abscissa_right = x - origin_right  # Distance in grid cells
        mask_right = abscissa_right >= 0

        # Normalized distance (both in grid cells)
        abscissa_norm_right = torch.clamp(abscissa_right / pml_width[1], 0, 1)

        sigma_right = sigma0 * (abscissa_norm_right**n_power)
        k_right = 1.0 + (k_max_cpml - 1.0) * (abscissa_norm_right**n_power)
        alpha_right = (
            alpha_max_cpml * (1.0 - abscissa_norm_right) + 0.1 * alpha_max_cpml
        )

        k = torch.where(mask_right, k_right, k)

        # b = exp(-(sigma/k + alpha) * dt / epsilon0)
        b_right = torch.exp(-(sigma_right / k_right + alpha_right) * dt / EP0)
        b = torch.where(mask_right, b_right, b)

        # a = sigma * (b - 1) / (k * (sigma + k * alpha))
        denom_right = k_right * (sigma_right + k_right * alpha_right) + eps
        a_right = sigma_right * (b_right - 1.0) / denom_right
        a_right = torch.where(sigma_right > 1e-6, a_right, torch.zeros_like(a_right))
        a = torch.where(mask_right, a_right, a)

    return a, b, k



"""Grid-related helpers for padding and boundary bookkeeping."""

from typing import Sequence, Union


def _normalize_grid_spacing_2d(
    grid_spacing: Union[float, Sequence[float]],
) -> list[float]:
    """Normalize 2D grid spacing to [dy, dx]."""
    if isinstance(grid_spacing, (int, float)):
        return [float(grid_spacing), float(grid_spacing)]
    return list(grid_spacing)


def _normalize_pml_width_2d(
    pml_width: Union[int, Sequence[int]],
) -> list[int]:
    """Normalize 2D PML width to [top, bottom, left, right]."""
    if isinstance(pml_width, int):
        return [pml_width] * 4
    pml_width_list = list(pml_width)
    if len(pml_width_list) == 1:
        return pml_width_list * 4
    if len(pml_width_list) == 2:
        return [
            pml_width_list[0],
            pml_width_list[0],
            pml_width_list[1],
            pml_width_list[1],
        ]
    return pml_width_list



import math

import pytest
import scipy.special
import torch

import tide


def analytic_trace_const_medium(
    wavelet: torch.Tensor,
    dt: float,
    src_pos_m: tuple[float, float],
    rec_pos_m: tuple[float, float],
    eps_r: float,
    sigma: float,
    current: float = 1.0,
) -> tuple[torch.Tensor, torch.Tensor]:
    """Analytical 2D TM solution in a homogeneous medium."""
    device = wavelet.device
    dtype = torch.float64
    nt = wavelet.numel()

    eps0 = 1.0 / (36.0 * math.pi) * 1e-9
    mu0 = 4.0 * math.pi * 1e-7

    t = torch.arange(nt, device=device, dtype=dtype) * dt
    r = (
        torch.tensor(rec_pos_m, device=device, dtype=dtype)
        - torch.tensor(src_pos_m, device=device, dtype=dtype)
    )
    R = torch.linalg.norm(r) + 1e-12

    ricker_real = wavelet.to(dtype)
    spectrum = torch.fft.rfft(ricker_real)

    freqs = torch.fft.rfftfreq(nt, d=dt).to(device)
    omega = 2.0 * math.pi * freqs
    omega_c = omega.to(torch.complex128)
    omega_safe = omega_c.clone()
    if omega_safe.numel() > 1:
        omega_safe[0] = omega_safe[1]
    else:
        omega_safe[0] = 1.0 + 0.0j

    eps_complex = (
        eps0 * torch.tensor(eps_r, device=device, dtype=torch.complex128)
        - 1j * torch.tensor(sigma, device=device, dtype=torch.float64) / omega_safe
    )
    k = omega_safe * torch.sqrt(mu0 * eps_complex)
    hankel0 = torch.from_numpy(
        scipy.special.hankel2(0, (k * R).cpu().numpy())
    ).to(device=device, dtype=torch.complex128)

    green = -current * omega_safe * mu0 * hankel0 / 4.0
    green[0] = 0.0 + 0.0j

    u_freq = spectrum * green
    u_time = torch.fft.irfft(u_freq, n=nt).real
    return t, u_time


def test_maxwelltm_matches_constant_medium_analytic():
    if not torch.cuda.is_available():
        pytest.skip("CUDA is required for Maxwell analytic test.")
    device = torch.device("cuda")
    dtype = torch.float64

    freq0 = 9e8  # Hz
    dt = 1e-11  # s
    nt = 800

    dx = dy = 0.005  # m
    eps_r = 10.0
    conductivity = 1e-3  # S/m

    ny, nx = 96, 128
    src_idx = (ny // 2, nx // 2)
    rec_idx = (ny // 2, nx // 2 + 20)  # ~0.1 m offset

    epsilon = torch.full((ny, nx), eps_r, device=device, dtype=dtype)
    sigma = torch.full_like(epsilon, conductivity)
    mu = torch.ones_like(epsilon)

    wavelet = tide.ricker(freq0, nt, dt, peak_time=1.0 / freq0, dtype=dtype, device=device)
    source_amplitude = wavelet.view(1, 1, nt)

    source_location = torch.tensor([[src_idx]], device=device)
    receiver_location = torch.tensor([[rec_idx]], device=device)

    _, _, _, _, _, _, _, receivers = tide.maxwelltm(
        epsilon,
        sigma,
        mu,
        grid_spacing=dy,
        dt=dt,
        source_amplitude=source_amplitude,
        source_location=source_location,
        receiver_location=receiver_location,
        stencil=2,
        pml_width=10,
        save_snapshots=False,

    )

    simulated = receivers[:, 0, 0].cpu()

    src_pos_m = (src_idx[0] * dy, src_idx[1] * dx)
    rec_pos_m = (rec_idx[0] * dy, rec_idx[1] * dx)

    _, analytic = analytic_trace_const_medium(
        wavelet=wavelet.cpu(),
        dt=dt,
        src_pos_m=src_pos_m,
        rec_pos_m=rec_pos_m,
        eps_r=eps_r,
        sigma=conductivity,
    )

    scale = torch.dot(simulated, analytic) / torch.dot(analytic, analytic)
    misfit = torch.linalg.norm(simulated - scale * analytic) / torch.linalg.norm(analytic)
    peak_shift = abs(int(simulated.abs().argmax()) - int(analytic.abs().argmax()))

    assert misfit < 0.05
    assert peak_shift <= 3



from typing import Union

import torch


def reverse_pad(pad: list[int]) -> list[int]:
    """Reverse the padding order for use with torch.nn.functional.pad.

    PyTorch's pad function expects padding in reverse order (last dim first).
    This function converts [y0, y1, x0, x1] to [x0, x1, y0, y1].

    Args:
        pad: Padding values in [y0, y1, x0, x1] format.

    Returns:
        Padding values in PyTorch format [x0, x1, y0, y1].
    """
    # For 2D: [y0, y1, x0, x1] -> [x0, x1, y0, y1]
    result = []
    for i in range(len(pad) // 2 - 1, -1, -1):
        result.extend([pad[i * 2], pad[i * 2 + 1]])
    return result


def create_or_pad(
    tensor: torch.Tensor,
    pad: Union[int, list[int]],
    device: torch.device,
    dtype: torch.dtype,
    size: tuple[int, ...],
    mode: str = "constant",
) -> torch.Tensor:
    """Creates a zero tensor of specified size or pads an existing tensor.

    If the input tensor is empty (numel == 0), a new zero tensor with the
    given size is created. Otherwise, the tensor is padded according to
    the specified mode.

    This is a unified padding function that supports:
    - Zero padding (mode='constant') for wavefields
    - Replicate padding (mode='replicate') for models

    Args:
        tensor: The input tensor to be created or padded.
        pad: The padding to apply. Can be an integer (for uniform padding)
            or a list of integers [y0, y1, x0, x1] for per-side padding.
        device: The PyTorch device for the tensor.
        dtype: The PyTorch data type for the tensor.
        size: The desired size of the tensor if it needs to be created.
        mode: Padding mode ('constant', 'replicate', 'reflect', 'circular').
            Default is 'constant' (zero padding)

    Returns:
        The created or padded tensor.

    Example:
        >>> # Create a zero tensor of size [2, 110, 110] (batch=2, with padding)
        >>> wf = create_or_pad(torch.empty(0), 5, device, dtype, (2, 110, 110))
        >>>
        >>> # Pad a wavefield with zeros [2, 100, 100] -> [2, 110, 110]
        >>> wf_padded = create_or_pad(wf, [5, 5, 5, 5], device, dtype, (2, 110, 110))
        >>>
        >>> # Pad a model with replicate mode [100, 100] -> [110, 110]
        >>> eps_padded = create_or_pad(eps, [5, 5, 5, 5], device, dtype, (110, 110), mode='replicate')
    """
    if isinstance(pad, int):
        # Convert single int to [pad, pad, pad, pad, ...] for each spatial dimension
        # size includes batch dimension if len > 2, so spatial ndim = len(size) - 1 or len(size)
        ndim = len(size) - 1 if len(size) > 2 else len(size)
        pad = [pad] * ndim * 2

    if tensor.numel() == 0:
        return torch.zeros(size, device=device, dtype=dtype)

    if max(pad) == 0:
        return tensor.clone()

    # Reverse padding for PyTorch's pad function
    reversed_pad = reverse_pad(pad)

    # For non-constant padding modes (replicate, reflect, circular),
    # PyTorch requires:
    # - 2D spatial padding: 3D or 4D input
    # - 3D spatial padding: 4D or 5D input
    original_ndim = tensor.ndim
    needs_unsqueeze = original_ndim in {2, 3} and mode != "constant"

    if needs_unsqueeze:
        tensor = tensor.unsqueeze(0)

    result = torch.nn.functional.pad(tensor, reversed_pad, mode=mode)

    if needs_unsqueeze:
        result = result.squeeze(0)

    # PyTorch's autograd system automatically tracks gradients through operations.
    # Explicitly calling requires_grad_() is incompatible with torch.func transforms.
    # Simply return the result; gradient tracking is handled automatically.
    return result


def zero_interior(
    tensor: torch.Tensor,
    fd_pad: Union[int, list[int]],
    pml_width: list[int],
    dim: int,
) -> torch.Tensor:
    """Zero out the interior region of a tensor (keeping only PML regions).

    This is used for CPML auxiliary variables which should only be non-zero
    in the PML regions. Setting the interior to zero allows the propagator
    to skip unnecessary PML calculations in those regions.

    Args:
        tensor: The input tensor with shape [batch, ny, nx].
        fd_pad: Finite difference padding. Can be an int or list [y0, y1, x0, x1].
        pml_width: The width of PML regions [top, bottom, left, right].
        dim: The spatial dimension to zero (0 for y, 1 for x).

    Returns:
        The tensor with interior region zeroed out.
    """
    shape = tensor.shape[1:]  # Spatial dimensions (without batch)
    ndim = len(shape)

    if isinstance(fd_pad, int):
        fd_pad = [fd_pad] * 2 * ndim

    # Calculate interior slice for the specified dimension
    interior_start = fd_pad[dim * 2] + pml_width[dim * 2]
    interior_end = shape[dim] - pml_width[dim * 2 + 1] - fd_pad[dim * 2 + 1]

    # Zero out the interior
    if dim == 0:  # y dimension
        tensor[:, interior_start:interior_end, :].fill_(0)
    else:  # x dimension
        tensor[:, :, interior_start:interior_end].fill_(0)

    return tensor



import math

import pytest

from tide.cfl import cfl_condition


def test_cfl_condition_warns_when_refining_dt():
    with pytest.warns(UserWarning):
        inner_dt, step_ratio = cfl_condition([0.1, 0.1], dt=0.1, max_vel=1.0)

    assert step_ratio >= 2
    assert math.isclose(inner_dt * step_ratio, 0.1)



import torch

from tide.callbacks import CallbackState, create_callback_state


def test_callback_state_views():
    ey = torch.arange(36, dtype=torch.float32).reshape(6, 6)
    models = {"epsilon": torch.ones_like(ey)}
    state = CallbackState(
        dt=0.1,
        step=2,
        nt=5,
        wavefields={"Ey": ey},
        models=models,
        fd_pad=[1, 1, 1, 1],
        pml_width=[1, 1, 1, 1],
    )

    torch.testing.assert_close(state.get_wavefield("Ey", view="full"), ey)
    torch.testing.assert_close(state.get_wavefield("Ey", view="pml"), ey[1:-1, 1:-1])
    torch.testing.assert_close(state.get_wavefield("Ey", view="inner"), ey[2:-2, 2:-2])
    torch.testing.assert_close(
        state.get_model("epsilon", view="inner"),
        models["epsilon"][2:-2, 2:-2],
    )


def test_create_callback_state_factory():
    ey = torch.zeros((2, 3), dtype=torch.float32)
    models = {"epsilon": torch.ones_like(ey)}
    gradients = {"epsilon": torch.full_like(ey, 2.0)}
    state = create_callback_state(
        dt=0.2,
        step=3,
        nt=10,
        wavefields={"Ey": ey},
        models=models,
        gradients=gradients,
        fd_pad=[1, 1, 1, 1],
        pml_width=[2, 2, 2, 2],
        is_backward=True,
        grid_spacing=[0.1, 0.1],
    )

    assert state.dt == 0.2
    assert state.step == 3
    assert state.nt == 10
    assert state.is_backward is True
    assert state.wavefield_names == ["Ey"]
    assert state.model_names == ["epsilon"]
    assert state.gradient_names == ["epsilon"]



import torch

from tide.padding import create_or_pad, reverse_pad, zero_interior


def test_reverse_pad_2d():
    assert reverse_pad([1, 2, 3, 4]) == [3, 4, 1, 2]


def test_create_or_pad_empty_and_constant():
    device = torch.device("cpu")
    dtype = torch.float32
    result = create_or_pad(torch.empty(0), 2, device, dtype, (2, 5, 6))
    assert result.shape == (2, 5, 6)
    assert torch.allclose(result, torch.zeros_like(result))

    base = torch.ones((2, 2), dtype=dtype, device=device)
    padded = create_or_pad(base, [1, 1, 1, 1], device, dtype, (4, 4))
    assert padded.shape == (4, 4)
    assert torch.allclose(padded[1:3, 1:3], base)
    assert padded[0, 0].item() == 0.0


def test_create_or_pad_replicate():
    device = torch.device("cpu")
    dtype = torch.float32
    base = torch.tensor([[1.0, 2.0], [3.0, 4.0]], device=device, dtype=dtype)
    padded = create_or_pad(base, [1, 1, 1, 1], device, dtype, (4, 4), mode="replicate")
    expected = torch.tensor(
        [
            [1.0, 1.0, 2.0, 2.0],
            [1.0, 1.0, 2.0, 2.0],
            [3.0, 3.0, 4.0, 4.0],
            [3.0, 3.0, 4.0, 4.0],
        ],
        device=device,
        dtype=dtype,
    )
    torch.testing.assert_close(padded, expected)


def test_zero_interior_y_and_x():
    tensor = torch.ones((1, 6, 6), dtype=torch.float32)
    fd_pad = [1, 1, 1, 1]
    pml_width = [1, 1, 1, 1]

    y_zeroed = zero_interior(tensor.clone(), fd_pad, pml_width, dim=0)
    assert torch.allclose(y_zeroed[:, 2:4, :], torch.zeros((1, 2, 6)))
    assert torch.all(y_zeroed[:, :2, :] == 1)
    assert torch.all(y_zeroed[:, 4:, :] == 1)

    x_zeroed = zero_interior(tensor.clone(), fd_pad, pml_width, dim=1)
    assert torch.allclose(x_zeroed[:, :, 2:4], torch.zeros((1, 6, 2)))
    assert torch.all(x_zeroed[:, :, :2] == 1)
    assert torch.all(x_zeroed[:, :, 4:] == 1)



import pytest
import torch


@pytest.mark.skipif(not torch.cuda.is_available(), reason="CUDA not available")
def test_fp8_storage_hw_only_sm89_roundtrip_and_grad():
    # This exercises the C/CUDA "with storage" path with FP8 compression.
    # It should not fall back to any software conversion and must work on SM89+.
    major, minor = torch.cuda.get_device_capability()
    if (major, minor) < (8, 9):
        pytest.skip(f"Requires SM89+, got {major}.{minor}")

    import tide

    device = torch.device("cuda")
    ny, nx = 64, 64
    dx = 0.02
    dt = 1e-11
    nt = 64

    epsilon = (torch.ones(ny, nx, device=device, dtype=torch.float32) * 4.0).requires_grad_(
        True
    )
    sigma = torch.zeros_like(epsilon)
    mu = torch.ones_like(epsilon)

    source_location = torch.tensor([[[ny // 2, nx // 2]]], device=device, dtype=torch.long)
    receiver_location = torch.tensor(
        [[[ny // 2, nx // 2 + 1]]], device=device, dtype=torch.long
    )
    wavelet = tide.ricker(2e8, nt, dt, peak_time=1.0 / 2e8).to(device)
    source_amplitude = wavelet.view(1, 1, nt)

    out = tide.maxwelltm(
        epsilon,
        sigma,
        mu,
        grid_spacing=dx,
        dt=dt,
        source_amplitude=source_amplitude,
        source_location=source_location,
        receiver_location=receiver_location,
        pml_width=10,
        save_snapshots=True,
        model_gradient_sampling_interval=2,
        storage_mode="device",
        storage_compression="fp8",
    )

    rec = out[-1]
    loss = (rec**2).mean()
    loss.backward()
    torch.cuda.synchronize()

    assert epsilon.grad is not None
    assert torch.isfinite(epsilon.grad).all()



"""Common electromagnetic wavelets for TIDE simulations.

This module provides various source wavelet functions commonly used in
electromagnetic wave simulations, particularly for Ground Penetrating Radar (GPR)
and other time-domain electromagnetic methods.

All wavelets return PyTorch tensors and support optional dtype specification.
"""

import math
from typing import Optional

import torch


def ricker(
    freq: float,
    length: int,
    dt: float,
    peak_time: Optional[float] = None,
    dtype: Optional[torch.dtype] = None,
    device: Optional[torch.device] = None,
) -> torch.Tensor:
    """Return a Ricker wavelet (Mexican hat wavelet).

    The Ricker wavelet is the negative normalized second derivative of a
    Gaussian function. It is commonly used in seismic and GPR simulations.

    The formula used is:
        w(t) = -(2*pi^2*(f*t' - 1)^2 - 1) * exp(-pi^2*(f*t' - 1)^2)

    where t' = t - peak_time.

    Args:
        freq: The central (dominant) frequency in Hz.
        length: The number of time samples.
        dt: The time sample spacing in seconds.
        peak_time: The time (in seconds) of the peak amplitude. If None,
            defaults to 1/freq (one period after start).
        dtype: The PyTorch datatype to use. Optional, defaults to float32.
        device: The PyTorch device to use. Optional, defaults to CPU.

    Returns:
        A PyTorch tensor representing the Ricker wavelet.

    Example:
        >>> # Create a 100 MHz Ricker wavelet for GPR
        >>> freq = 100e6  # 100 MHz
        >>> dt = 1e-10    # 0.1 ns time step
        >>> length = 500  # 500 time samples
        >>> wavelet = ricker(freq, length, dt)

    """
    if dt == 0:
        raise ValueError("dt cannot be zero.")
    if freq <= 0:
        raise ValueError("freq must be positive.")
    if length <= 0:
        raise ValueError("length must be positive.")

    if peak_time is None:
        peak_time = 1.0 / freq  # Default: one period

    t = torch.arange(float(length), dtype=dtype, device=device) * dt
    t_shifted = t - peak_time

    # Ricker wavelet formula
    pi2_f2_t2 = (math.pi * freq * t_shifted) ** 2
    y = (1 - 2 * pi2_f2_t2) * torch.exp(-pi2_f2_t2)

    return y




import math

import torch

from tide.resampling import downsample, downsample_and_movedim, upsample


def test_upsample_downsample_roundtrip_low_freq():
    device = torch.device("cpu")
    dtype = torch.float32
    step_ratio = 2
    n = 64
    t = torch.arange(n, device=device, dtype=dtype)
    signal = torch.sin(2.0 * math.pi * 4.0 * t / n)  # 4 cycles over length
    signal = signal[None, None, :]

    up = upsample(signal, step_ratio=step_ratio)
    down = downsample(up, step_ratio=step_ratio)
    torch.testing.assert_close(down, signal, atol=1e-4, rtol=1e-4)


def test_downsample_and_movedim_matches_manual():
    device = torch.device("cpu")
    dtype = torch.float32
    step_ratio = 2
    receiver = torch.randn(6, 2, 3, device=device, dtype=dtype)
    expected = downsample(torch.movedim(receiver, 0, -1), step_ratio=step_ratio)
    actual = downsample_and_movedim(receiver, step_ratio=step_ratio)
    torch.testing.assert_close(actual, expected)



"""Tests for staggered grid operations in tide.staggered module."""

import math

import pytest
import torch

from tide import staggered


class TestDiffY1:
    """Tests for diffy1 - first y derivative at integer grid points."""

    @pytest.mark.parametrize("stencil", [2, 4, 6, 8])
    def test_diffy1_linear_function(self, stencil):
        """First derivative of linear function should be constant."""
        ny, nx = 32, 32
        dy = 0.1
        rdy = torch.tensor(1.0 / dy, dtype=torch.float32)

        # f(y) = 2*y, so df/dy = 2
        y = torch.arange(ny, dtype=torch.float32) * dy
        f = 2.0 * y
        a = f.expand(nx, ny).T  # Shape: [ny, nx]

        result = staggered.diffy1(a, stencil, rdy)

        # The result should be approximately 2.0 in the valid region
        expected = torch.full((ny, nx), 2.0, dtype=torch.float32)

        # Valid region depends on stencil order:
        # stencil=2: [1:]  (1 row top padding)
        # stencil=4: [2:-1] (2 rows top, 1 row bottom)
        # stencil=6: [3:-2] (3 rows top, 2 rows bottom)
        # stencil=8: [4:-3] (4 rows top, 3 rows bottom)
        pad_top = stencil // 2
        pad_bottom = pad_top - 1 if stencil > 2 else 0
        actual = result[pad_top:ny-pad_bottom if pad_bottom > 0 else ny, :]
        expected_actual = expected[pad_top:ny-pad_bottom if pad_bottom > 0 else ny, :]

        torch.testing.assert_close(actual, expected_actual, atol=1e-5, rtol=1e-5)

    @pytest.mark.parametrize("stencil", [2, 4, 6, 8])
    def test_diffy1_sine_function(self, stencil):
        """First derivative of sin should be cos."""
        ny, nx = 64, 32
        dy = 0.05
        rdy = torch.tensor(1.0 / dy, dtype=torch.float32)

        y = torch.arange(ny, dtype=torch.float32) * dy
        f = torch.sin(2.0 * math.pi * y / (ny * dy))
        a = f.expand(nx, ny).T

        result = staggered.diffy1(a, stencil, rdy)

        # Expected: df/dy = (2*pi/L) * cos(2*pi*y/L)
        L = ny * dy
        k = 2.0 * math.pi / L
        expected = k * torch.cos(2.0 * math.pi * y / L)

        # Check in the valid region (away from boundaries)
        pad_top = stencil // 2 + 2
        pad_bottom = pad_top - 1 if stencil > 2 else 1
        actual = result[pad_top:ny-pad_bottom, 0]
        expected_slice = expected[pad_top:ny-pad_bottom]

        # Use absolute tolerance mainly, as relative tolerance can blow up near zeros
        # Higher stencil orders should have better accuracy
        atol = 0.15 if stencil == 2 else 0.12
        torch.testing.assert_close(actual, expected_slice, atol=atol, rtol=1.0)

    def test_diffy1_output_shape(self):
        """Output shape should match input shape."""
        ny, nx = 16, 24
        a = torch.randn(ny, nx, dtype=torch.float32)
        rdy = torch.tensor(1.0, dtype=torch.float32)

        for stencil in [2, 4, 6, 8]:
            result = staggered.diffy1(a, stencil, rdy)
            assert result.shape == a.shape


class TestDiffX1:
    """Tests for diffx1 - first x derivative at integer grid points."""

    @pytest.mark.parametrize("stencil", [2, 4, 6, 8])
    def test_diffx1_linear_function(self, stencil):
        """First derivative of linear function should be constant."""
        ny, nx = 32, 32
        dx = 0.1
        rdx = torch.tensor(1.0 / dx, dtype=torch.float32)

        # f(x) = 3*x, so df/dx = 3
        x = torch.arange(nx, dtype=torch.float32) * dx
        f = 3.0 * x
        a = f.expand(ny, nx)  # Shape: [ny, nx]

        result = staggered.diffx1(a, stencil, rdx)

        # The result should be approximately 3.0 in the valid region
        expected = torch.full((ny, nx), 3.0, dtype=torch.float32)

        # Valid region depends on stencil order
        pad_left = stencil // 2
        pad_right = pad_left - 1 if stencil > 2 else 0
        actual = result[:, pad_left:nx-pad_right if pad_right > 0 else nx]
        expected_actual = expected[:, pad_left:nx-pad_right if pad_right > 0 else nx]

        torch.testing.assert_close(actual, expected_actual, atol=1e-5, rtol=1e-5)

    @pytest.mark.parametrize("stencil", [2, 4, 6, 8])
    def test_diffx1_sine_function(self, stencil):
        """First derivative of sin should be cos."""
        ny, nx = 32, 64
        dx = 0.05
        rdx = torch.tensor(1.0 / dx, dtype=torch.float32)

        x = torch.arange(nx, dtype=torch.float32) * dx
        f = torch.sin(2.0 * math.pi * x / (nx * dx))
        a = f.expand(ny, nx)

        result = staggered.diffx1(a, stencil, rdx)

        # Expected: df/dx = (2*pi/L) * cos(2*pi*x/L)
        L = nx * dx
        k = 2.0 * math.pi / L
        expected = k * torch.cos(2.0 * math.pi * x / L)

        pad_left = stencil // 2 + 2
        pad_right = pad_left - 1 if stencil > 2 else 1
        actual = result[0, pad_left:nx-pad_right]
        expected_slice = expected[pad_left:nx-pad_right]

        # Use absolute tolerance mainly, as relative tolerance can blow up near zeros
        atol = 0.15 if stencil == 2 else 0.12
        torch.testing.assert_close(actual, expected_slice, atol=atol, rtol=1.0)

    def test_diffx1_output_shape(self):
        """Output shape should match input shape."""
        ny, nx = 16, 24
        a = torch.randn(ny, nx, dtype=torch.float32)
        rdx = torch.tensor(1.0, dtype=torch.float32)

        for stencil in [2, 4, 6, 8]:
            result = staggered.diffx1(a, stencil, rdx)
            assert result.shape == a.shape


class TestDiffYH1:
    """Tests for diffyh1 - first y derivative at half grid points."""

    @pytest.mark.parametrize("stencil", [2, 4, 6, 8])
    def test_diffyh1_linear_function(self, stencil):
        """First derivative of linear function should be constant."""
        ny, nx = 32, 32
        dy = 0.1
        rdy = torch.tensor(1.0 / dy, dtype=torch.float32)

        y = torch.arange(ny, dtype=torch.float32) * dy
        f = 2.0 * y
        a = f.expand(nx, ny).T

        result = staggered.diffyh1(a, stencil, rdy)

        expected = torch.full((ny, nx), 2.0, dtype=torch.float32)

        # For half-grid derivatives, valid region is:
        # stencil=2: [1:-1]  (1 row top, 1 row bottom)
        # stencil=4: [2:-2] (2 rows top, 2 rows bottom)
        # etc.
        pad = stencil // 2
        actual = result[pad:ny-pad, :]
        expected_actual = expected[pad:ny-pad, :]

        torch.testing.assert_close(actual, expected_actual, atol=1e-5, rtol=1e-5)

    def test_diffyh1_output_shape(self):
        """Output shape should match input shape."""
        ny, nx = 16, 24
        a = torch.randn(ny, nx, dtype=torch.float32)
        rdy = torch.tensor(1.0, dtype=torch.float32)

        for stencil in [2, 4, 6, 8]:
            result = staggered.diffyh1(a, stencil, rdy)
            assert result.shape == a.shape


class TestDiffXH1:
    """Tests for diffxh1 - first x derivative at half grid points."""

    @pytest.mark.parametrize("stencil", [2, 4, 6, 8])
    def test_diffxh1_linear_function(self, stencil):
        """First derivative of linear function should be constant."""
        ny, nx = 32, 32
        dx = 0.1
        rdx = torch.tensor(1.0 / dx, dtype=torch.float32)

        x = torch.arange(nx, dtype=torch.float32) * dx
        f = 3.0 * x
        a = f.expand(ny, nx)

        result = staggered.diffxh1(a, stencil, rdx)

        expected = torch.full((ny, nx), 3.0, dtype=torch.float32)

        # For half-grid derivatives, valid region is symmetric
        pad = stencil // 2
        actual = result[:, pad:nx-pad]
        expected_actual = expected[:, pad:nx-pad]

        torch.testing.assert_close(actual, expected_actual, atol=1e-5, rtol=1e-5)

    def test_diffxh1_output_shape(self):
        """Output shape should match input shape."""
        ny, nx = 16, 24
        a = torch.randn(ny, nx, dtype=torch.float32)
        rdx = torch.tensor(1.0, dtype=torch.float32)

        for stencil in [2, 4, 6, 8]:
            result = staggered.diffxh1(a, stencil, rdx)
            assert result.shape == a.shape


class TestDiffZH1:
    """Tests for diffzh1 - first z derivative at half grid points (3D)."""

    @pytest.mark.parametrize("stencil", [2, 4, 6, 8])
    def test_diffzh1_linear_function(self, stencil):
        """First derivative of linear function should be constant."""
        nz, ny, nx = 16, 16, 16
        dz = 0.1
        rdz = torch.tensor(1.0 / dz, dtype=torch.float32)

        z = torch.arange(nz, dtype=torch.float32) * dz
        f = 2.0 * z
        a = f.reshape(nz, 1, 1).expand(nz, ny, nx)

        result = staggered.diffzh1(a, stencil, rdz)

        expected = torch.full((nz, ny, nx), 2.0, dtype=torch.float32)

        # For half-grid derivatives, valid region is symmetric
        pad = stencil // 2
        actual = result[pad:nz-pad, :, :]
        expected_actual = expected[pad:nz-pad, :, :]

        torch.testing.assert_close(actual, expected_actual, atol=1e-5, rtol=1e-5)

    def test_diffzh1_output_shape(self):
        """Output shape should match input shape."""
        nz, ny, nx = 8, 12, 16
        a = torch.randn(nz, ny, nx, dtype=torch.float32)
        rdz = torch.tensor(1.0, dtype=torch.float32)

        for stencil in [2, 4, 6, 8]:
            result = staggered.diffzh1(a, stencil, rdz)
            assert result.shape == a.shape


class TestSetPMLProfiles2D:
    """Tests for set_pml_profiles - 2D PML profile setup."""

    def test_set_pml_profiles_2d_shapes(self):
        """PML profiles should have correct shapes."""
        pml_width = [4, 4, 4, 4]  # [y0, y1, x0, x1]
        accuracy = 4
        fd_pad = [2, 2, 2, 2]
        dt = 1e-11
        grid_spacing = [0.01, 0.01]
        max_vel = 3e8
        dtype = torch.float32
        device = torch.device("cpu")
        pml_freq = 25.0
        ny, nx = 32, 32

        profiles = staggered.set_pml_profiles(
            pml_width, accuracy, fd_pad, dt, grid_spacing,
            max_vel, dtype, device, pml_freq, ny, nx
        )
        ab_profiles, k_profiles = profiles[:8], profiles[8:]

        # ab_profiles: [ay, ayh, ax, axh, by, byh, bx, bxh]
        # k_profiles: [ky, kyh, kx, kxh]
        assert len(ab_profiles) == 8
        assert len(k_profiles) == 4

        # Each profile should have length matching the dimension
        assert ab_profiles[0].shape == (1, ny, 1)  # ay
        assert ab_profiles[2].shape == (1, 1, nx)  # ax
        assert k_profiles[0].shape == (1, ny, 1)   # ky
        assert k_profiles[2].shape == (1, 1, nx)   # kx

    def test_set_pml_profiles_zero_width(self):
        """Zero PML width should produce valid (mostly zero) profiles."""
        pml_width = [0, 0, 0, 0]
        accuracy = 2
        fd_pad = [1, 1, 1, 1]
        dt = 1e-11
        grid_spacing = [0.01, 0.01]
        max_vel = 3e8
        dtype = torch.float32
        device = torch.device("cpu")
        pml_freq = 25.0
        ny, nx = 16, 16

        profiles = staggered.set_pml_profiles(
            pml_width, accuracy, fd_pad, dt, grid_spacing,
            max_vel, dtype, device, pml_freq, ny, nx
        )
        ab_profiles, k_profiles = profiles[:8], profiles[8:]

        # With zero PML width, a and b profiles should be all zeros
        # k profiles should be all ones
        for ab in ab_profiles:
            assert torch.all(ab == 0.0)
        for k in k_profiles:
            assert torch.all(k == 1.0)

    def test_set_pml_profiles_coefficient_ranges(self):
        """PML coefficients should be in physically reasonable ranges."""
        pml_width = [8, 8, 8, 8]
        accuracy = 4
        fd_pad = [2, 2, 2, 2]
        dt = 1e-11
        grid_spacing = [0.01, 0.01]
        max_vel = 3e8
        dtype = torch.float32
        device = torch.device("cpu")
        pml_freq = 25.0
        ny, nx = 32, 32
        profiles = staggered.set_pml_profiles(
            pml_width, accuracy, fd_pad, dt, grid_spacing,
            max_vel, dtype, device, pml_freq, ny, nx
        )
        
        ay, ayh, ax, axh, by, byh, bx, bxh,ky, kyh, kx, kxh = profiles
        # a coefficients can be negative or small (CPML recursive convolution)
        # b coefficients are decay factors, should be in [0, 1]
        for b in [by, byh, bx, bxh]:
            assert torch.all(b >= 0.0), "b coefficients should be non-negative"
            assert torch.all(b <= 1.0), "b coefficients should be <= 1"

        # k coefficients should be >= 1 (stretching factor)
        for k in [ky, kyh, kx, kxh]:
            assert torch.all(k >= 1.0), "k profiles should be >= 1"
            # k should be 1 in interior and > 1 in PML
            assert k.max() > 1.0, "k should be > 1 in PML region"
